==> web2sms_0.1.0.tar.gz: web2sms-0.1.0/src/providers/o2_germany/o2_germany.cpp <==
/*
 *
 *  Copyright (c) 2010 Christoph Keller <gri@nospam@not-censored.com>
 *
 *  This file is part of Web2SMS.
 *
 *  Web2SMS is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Web2SMS is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Web2SMS. If not, see <http://www.gnu.org/licenses/>
 *
 */

// Local includes
#include "o2_germany.hpp"
#include "../accountsettingsdialog.hpp"

// Global includes
#include <QtCore/QRegExp>
#include <QtCore/QDebug>
#include <QtCore/QSettings>
#include <QtCore/QFile>
#include <QtNetwork/QNetworkAccessManager>
#include <QtNetwork/QNetworkRequest>
#include <QtNetwork/QNetworkReply>

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

class o2GermanyPrivate : public QObject
{
  Q_OBJECT

public:
  o2GermanyPrivate()
  : sentMessages(0)
  , freeSMSCount(-1)
  , settings(QSettings::UserScope, "maemo", "web2sms")
  {
    settings.beginGroup("providers/o2_germany");

    // We define some default urls (may be replaced by code later)
    urls.preLoginUrl      = QUrl("http://www.o2online.de");
    urls.logoutUrl        = QUrl("https://login.o2online.de/loginRegistration/loginAction.do?_flowId=logout");
    urls.smsCenterUrl     = QUrl("https://email.o2online.de/smscenter_new.osp");
    urls.smsCenterSendUrl = QUrl("https://email.o2online.de/smscenter_send.osp");

    // Read the settings
    loginName = settings.value("loginName").toString();
    password  = settings.value("password").toString();

    connect(&accessManager, SIGNAL(finished(QNetworkReply*)), this, SLOT(onReplyFinished(QNetworkReply*)));
  }

  enum State
  {
    Unknown,
    PreLogin,
    Login,
    Logout,
    ComCenter,
    SmsCenter,
    SmsCenterSend
  };

  o2Germany* self;

  // Properties
  QNetworkAccessManager accessManager;

  struct
  {
    QUrl preLoginUrl;
    QUrl loginUrl;
    QUrl logoutUrl;

    QUrl comCenterUrl;
    QUrl smsCenterUrl;
    QUrl smsCenterSendUrl;
  } urls;

  QHash<QNetworkReply*, State> replyStateHash;

  QSettings settings;

  // User settings
  QString loginName;
  QString password;

  int sentMessages;
  int freeSMSCount;

  QHash<QString, QByteArray> sendFormDefaultValues;

  // Functions
  QNetworkRequest prepareRequest() const;
  
  // Network functions
  QNetworkReply* get(const QNetworkRequest& request, State state);
  QNetworkReply* post(const QNetworkRequest& request, const QByteArray& data, State state);

  // Handler functions
  void handlePreLogin(QNetworkReply* reply);
  void handleLogin(QNetworkReply* reply);
  void handleComCenter(QNetworkReply* reply);
  void handleSmsCenter(QNetworkReply* reply);
  void handleSmsCenterSend(QNetworkReply* reply);
  void handleLogout(QNetworkReply* reply);

public slots:
  void onReplyFinished(QNetworkReply* reply);
};

//////////////////////////////////////////////////////////////////////////

QNetworkRequest o2GermanyPrivate::prepareRequest() const
{
  QNetworkRequest request;
  request.setRawHeader("User-Agent", "Opera/9.80 (Linux x86_64; U; en) Presto/2.5.18 Version/10.50 Gentoo");

  return request;
}

//////////////////////////////////////////////////////////////////////////

QNetworkReply* o2GermanyPrivate::get(const QNetworkRequest& request, State state)
{
  // Get the request
  QNetworkReply* reply = accessManager.get(request);

  // Remember the reply with the given state
  replyStateHash.insert(reply, state);

  return reply;
}

//////////////////////////////////////////////////////////////////////////

QNetworkReply* o2GermanyPrivate::post(const QNetworkRequest& request, const QByteArray& data, State state)
{
  // Get the post
  QNetworkReply* reply = accessManager.post(request, data);

  // Remember the reply with the given state
  replyStateHash.insert(reply, state);

  return reply;
}

//////////////////////////////////////////////////////////////////////////

void o2GermanyPrivate::handlePreLogin(QNetworkReply* reply)
{
  // Get all data from the reply
  QByteArray readData = reply->readAll();
  QString data = QString::fromUtf8(readData.constData(), readData.size());

  // Find the login url
  QRegExp formExp("loginForm\" action=\"([^\"]*)\"");
  if ( formExp.indexIn(data) != -1 )
  {
    // Set the login url
    urls.loginUrl = QUrl(formExp.cap(1));

    QRegExp urlExp("<option value=\"([^\"]*comcenter[^\"]*)\"");
    urlExp.indexIn(data, formExp.pos() + formExp.matchedLength());

    urls.comCenterUrl = urlExp.cap(1);

    // Ok, the fields are known:
    // loginName
    // password
    // url

    QByteArray formData;
    formData += "loginName=" + loginName.toUtf8();
    formData += "&password=" + password.toUtf8();
    formData += "&url=" + urlExp.cap(1).toUtf8().toPercentEncoding();

    // Prepare a request and post it
    QNetworkRequest request = prepareRequest();
    request.setRawHeader("Referer", reply->url().toEncoded());
    request.setUrl( urls.loginUrl );

    post(request, formData, Login);
  }
  else
  {
    // Emit that we did not find the loginForm
    emit self->error( tr("loginForm could not be found!") );

    // Emit that the login failed
    emit self->loginReply(false);
  }
}

//////////////////////////////////////////////////////////////////////////

void o2GermanyPrivate::handleLogin(QNetworkReply* reply)
{
  // Get all data from the reply
  QByteArray readData = reply->readAll();
  QString data = QString::fromUtf8(readData.constData(), readData.size());

  bool success = data.indexOf("Fehler beim Login") == -1;

  if ( success )
  {
    // Open a new request to the sms center
    QNetworkRequest request = prepareRequest();
    request.setRawHeader("Referer", reply->url().toEncoded());
    request.setUrl( urls.comCenterUrl );
    get(request, ComCenter);
  }

  // Emit that the login failed
  emit self->loginReply( success );
}

//////////////////////////////////////////////////////////////////////////

void o2GermanyPrivate::handleComCenter(QNetworkReply* reply)
{
  // Get all data from the reply
  QByteArray readData = reply->readAll();

  QNetworkRequest request = prepareRequest();
  request.setRawHeader("Referer", reply->url().toEncoded());
  request.setUrl( urls.smsCenterUrl );
  get(request, SmsCenter);
}

//////////////////////////////////////////////////////////////////////////

void o2GermanyPrivate::handleSmsCenter(QNetworkReply* reply)
{
  // Get all data from the reply
  QByteArray readData = reply->readAll();
  QString data = QString::fromUtf8(readData.constData(), readData.size());

  int formOffset = 0;

  // Find the message which tells us something about our remaining free-sms
  // If it's not found we don't care ...
  QRegExp freeSmsExp("Frei-SMS: (\\d*) Web2SMS");
  if ( freeSmsExp.indexIn(data) != -1 )
  {
    // Remember the offset to not search from the beginning every time
    formOffset = freeSmsExp.pos() + freeSmsExp.matchedLength();

    // How many free sms are left?
    freeSMSCount = freeSmsExp.cap(1).toInt();

    // Emit how many free-sms we have left
    emit self->freeSmsCountChanged( freeSMSCount );
  }

  // Try to determine how many free sms we have left
  QRegExp formExp("name=\"frmSMS\" action=\"([^\"]*)\"");
  if ( formExp.indexIn(data, formOffset) != -1 )
  {
    // Remember the offset
    int offset = formExp.pos() + formExp.matchedLength();

    // Change the free sms center send-url if wanted
    urls.smsCenterSendUrl = reply->url().resolved(formExp.cap(1));

    // Fill the default fields
    sendFormDefaultValues.clear();

    QRegExp fieldExp("<input type=\"hidden\" name=\"([^\"]*)\" value=\"([^\"]*)\"", Qt::CaseInsensitive);
    int pos = fieldExp.indexIn(data, offset);
    while ( pos != -1 )
    {
      sendFormDefaultValues.insert( fieldExp.cap(1), fieldExp.cap(2).toUtf8().toPercentEncoding() );
      
      // Find the next entry
      pos = fieldExp.indexIn(data, pos+fieldExp.matchedLength());
    }

    sendFormDefaultValues.insert("Frequency", "5");
    sendFormDefaultValues.insert("StartDateDay", QByteArray());
    sendFormDefaultValues.insert("StartDateMonth", QByteArray());
    sendFormDefaultValues.insert("StartDateYear", QByteArray());
    sendFormDefaultValues.insert("StartDateHour", QByteArray());
    sendFormDefaultValues.insert("StartDateMin", QByteArray());
    sendFormDefaultValues.insert("EndDateDay", QByteArray());
    sendFormDefaultValues.insert("EndDateMonth", QByteArray());
    sendFormDefaultValues.insert("EndDateYear", QByteArray());
    sendFormDefaultValues.insert("EndDateHour", QByteArray());
    sendFormDefaultValues.insert("EndDateMin", QByteArray());
  }
  else
  {
#ifdef Q_WS_MAEMO_5
    QFile file("/home/user/MyDocs/o2sms_debug_handleSmsCenter.html");
    if ( file.open(QIODevice::Text | QIODevice::WriteOnly) )
    {
      file.write(readData);
      file.close();
    }
#endif // Q_WS_MAEMO5

    // Emit that we could not find the frmSMS
    emit self->error( tr("frmSMS could not be found!") );
  }
}

//////////////////////////////////////////////////////////////////////////

void o2GermanyPrivate::handleSmsCenterSend(QNetworkReply* reply)
{
  // Get all data from the reply
  QByteArray readData = reply->readAll();
  QString data = QString::fromUtf8(readData.constData(), readData.size());

  // TODO: ERROR CHECKING!!!!

  // Emit a signal that we've sent the message
  emit self->sendMessageReply(true, sentMessages);

  // Emit a signal that the free sms count may have changed?
  emit self->freeSmsCountChanged( --freeSMSCount );
}

//////////////////////////////////////////////////////////////////////////

void o2GermanyPrivate::handleLogout(QNetworkReply* reply)
{
  Q_UNUSED(reply);

  // No more free sms
  freeSMSCount = -1;

  // No need to read something, logout works always
  emit self->logoutReply(true);
}

//////////////////////////////////////////////////////////////////////////

void o2GermanyPrivate::onReplyFinished(QNetworkReply* reply)
{
  // Handle a redirection if needed
  QVariant redirectLocation = reply->attribute(QNetworkRequest::RedirectionTargetAttribute);
  if ( !redirectLocation.isNull() )
  {
    // Where are we going to?
    QUrl newUrl = reply->url().resolved(redirectLocation.toUrl());

    // Prepare the request and send it
    QNetworkRequest request = prepareRequest();
    request.setRawHeader("Referer", reply->url().toEncoded());
    request.setUrl( newUrl );
    get(request, replyStateHash.value(reply));

    return; // Don't continue
  }

  // If some error happend while transmitting HTTP content, also signalize it
  if ( reply->error() )
    emit self->error( tr("HTTP error happened:\n%1").arg(reply->errorString()) );

  // Decide what to do now
  switch(replyStateHash.value(reply))
  {
  case PreLogin:
    handlePreLogin(reply);
    break;

  case Login:
    handleLogin(reply);
    break;

  case ComCenter:
    handleComCenter(reply);
    break;

  case SmsCenter:
    handleSmsCenter(reply);
    break;

  case SmsCenterSend:
    handleSmsCenterSend(reply);
    break;

  case Logout:
    handleLogout(reply);
    break;

  default:
    break;
  }

  // Remove hash values for this reply (if any)
  replyStateHash.remove(reply);

  // Delete the reply at a later time
  reply->deleteLater();
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

o2Germany::o2Germany(QObject* parent /* = 0 */)
: ProviderInterface(parent)
, d(new o2GermanyPrivate)
{
  d->self = this;
}

//////////////////////////////////////////////////////////////////////////

o2Germany::~o2Germany()
{
  delete d;
}

//////////////////////////////////////////////////////////////////////////

QString o2Germany::name() const
{
  return tr("o2 Germany");
}

//////////////////////////////////////////////////////////////////////////

QPixmap o2Germany::pixmap() const
{
  return QPixmap(":/providers/o2_germany/o2_germany.png");
}

//////////////////////////////////////////////////////////////////////////

bool o2Germany::showAccountSettingsDialog(QWidget* parent)
{
  AccountSettingsDialog dialog(parent);
  dialog.setWindowTitle( name() );
  dialog.setUserName(d->loginName);
  dialog.setPassword(d->password);

  if ( dialog.exec() == QDialog::Accepted )
  {
    d->loginName = dialog.userName();
    d->password  = dialog.password();

    d->settings.setValue("loginName", d->loginName);
    d->settings.setValue("password", d->password);

    return true;
  }

  return false;
}

//////////////////////////////////////////////////////////////////////////

bool o2Germany::hasLoginData() const
{
  return !d->loginName.isEmpty();
}

//////////////////////////////////////////////////////////////////////////

bool o2Germany::isLoggedIn() const
{
  // Simple .. heh?
  return d->freeSMSCount != -1;
}

//////////////////////////////////////////////////////////////////////////

int o2Germany::freeSMSCount() const
{
  return d->freeSMSCount;
}

//////////////////////////////////////////////////////////////////////////

void o2Germany::login()
{
  // Start the first request
  QNetworkRequest request = d->prepareRequest();
  request.setUrl( d->urls.preLoginUrl );

  // Start the pre-login
  d->get(request, o2GermanyPrivate::PreLogin);
}

//////////////////////////////////////////////////////////////////////////

void o2Germany::logout()
{
  // Start the logout request
  QNetworkRequest request = d->prepareRequest();
  request.setUrl( d->urls.logoutUrl );

  // Start the logout request
  d->get(request, o2GermanyPrivate::Logout);
}

//////////////////////////////////////////////////////////////////////////

int o2Germany::sendMessage(const QString& receiver, const QString& message)
{
  // Prepare the send-message request
  QNetworkRequest request = d->prepareRequest();
  request.setRawHeader("Referer", d->urls.smsCenterUrl.toEncoded());
  request.setHeader(QNetworkRequest::ContentTypeHeader, "application/x-www-form-urlencoded");
  request.setUrl( d->urls.smsCenterSendUrl );

  qDebug() << d->urls.smsCenterSendUrl;

  // Build the form data
  QByteArray formData;

  // Fill the remaining fields
  d->sendFormDefaultValues.insert("SMSTo", receiver.toUtf8().toPercentEncoding());
  d->sendFormDefaultValues.insert("SMSText", message.toUtf8().toPercentEncoding());

  // Fill the default values
  QHashIterator<QString, QByteArray> iter(d->sendFormDefaultValues);
  while (iter.hasNext())
  {
    iter.next();

    if ( !formData.isEmpty() )
      formData += '&';

    formData += iter.key().toUtf8() + '=' + iter.value();
  }

  d->post(request, formData, o2GermanyPrivate::SmsCenterSend);

  // Return a new message id
  return d->sentMessages++;
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

#include "o2_germany.moc"

==> web2sms_0.1.1-4.tar.gz: web2sms-0.1.1/src/providers/o2_germany/o2_germany.cpp <==
/*
 *
 *  Copyright (c) 2010 Christoph Keller <gri@nospam@not-censored.com>
 *
 *  This file is part of Web2SMS.
 *
 *  Web2SMS is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Web2SMS is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Web2SMS. If not, see <http://www.gnu.org/licenses/>
 *
 */

// Local includes
#include "o2_germany.hpp"
#include "../providerplugin.hpp"
#include "../accountsettingsdialog.hpp"

// Global includes
#include <QtCore/QRegExp>
#include <QtCore/QDebug> 
#include <QtCore/QDataStream>
#include <QtCore/QFile>
#include <QtNetwork/QNetworkAccessManager>
#include <QtNetwork/QNetworkRequest>
#include <QtNetwork/QNetworkReply>

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

class o2GermanyPrivate : public QObject
{
  Q_OBJECT

public:
  o2GermanyPrivate()
  : sentMessages(0)
  , freeSMSCount(-1)
  {
    // We define some default urls (may be replaced by code later)
    urls.preLoginUrl      = QUrl("http://www.o2online.de");
    urls.logoutUrl        = QUrl("https://login.o2online.de/loginRegistration/loginAction.do?_flowId=logout");
    urls.smsCenterUrl     = QUrl("https://email.o2online.de/smscenter_new.osp");
    urls.smsCenterSendUrl = QUrl("https://email.o2online.de/smscenter_send.osp");
    urls.comCenterUrl     = QUrl("https://email.o2online.de/ssomanager.osp?APIID=AUTH-WEBSSO&TargetApp=m_folder_detail.osp%3FFolder%3DO3sis%3A%3AInbox");

    // Read the settings
    //loginName = settings.value("loginName").toString();
    //password  = settings.value("password").toString();

    connect(&accessManager, SIGNAL(finished(QNetworkReply*)), this, SLOT(onReplyFinished(QNetworkReply*)));
  }

  enum State
  {
    Unknown,
    PreLogin,
    Login,
    Logout,
    ComCenter,
    SmsCenter,
    SmsCenterSend
  };

  o2Germany* self;

  // Properties
  QNetworkAccessManager accessManager;

  struct
  {
    QUrl preLoginUrl;
    QUrl loginUrl;
    QUrl logoutUrl;

    QUrl comCenterUrl;
    QUrl smsCenterUrl;
    QUrl smsCenterSendUrl;
  } urls;

  QHash<QNetworkReply*, State> replyStateHash;

  // User settings
  QString loginName;
  QString password;

  int sentMessages;
  int freeSMSCount;

  QHash<QString, QByteArray> sendFormDefaultValues;

  // Functions
  QNetworkRequest prepareRequest() const;
  
  // Network functions
  QNetworkReply* get(const QNetworkRequest& request, State state);
  QNetworkReply* post(const QNetworkRequest& request, const QByteArray& data, State state);

  // Handler functions
  void handlePreLogin(QNetworkReply* reply);
  void handleLogin(QNetworkReply* reply);
  void handleComCenter(QNetworkReply* reply);
  void handleSmsCenter(QNetworkReply* reply);
  void handleSmsCenterSend(QNetworkReply* reply);
  void handleLogout(QNetworkReply* reply);

public slots:
  void onReplyFinished(QNetworkReply* reply);
};

//////////////////////////////////////////////////////////////////////////

QNetworkRequest o2GermanyPrivate::prepareRequest() const
{
  QNetworkRequest request;
  request.setRawHeader("User-Agent", "Opera/9.80 (Linux x86_64; U; en) Presto/2.5.18 Version/10.50 Gentoo");

  return request;
}

//////////////////////////////////////////////////////////////////////////

QNetworkReply* o2GermanyPrivate::get(const QNetworkRequest& request, State state)
{
  // Get the request
  QNetworkReply* reply = accessManager.get(request);

  // Remember the reply with the given state
  replyStateHash.insert(reply, state);

  return reply;
}

//////////////////////////////////////////////////////////////////////////

QNetworkReply* o2GermanyPrivate::post(const QNetworkRequest& request, const QByteArray& data, State state)
{
  // Get the post
  QNetworkReply* reply = accessManager.post(request, data);

  // Remember the reply with the given state
  replyStateHash.insert(reply, state);

  return reply;
}

//////////////////////////////////////////////////////////////////////////

void o2GermanyPrivate::handlePreLogin(QNetworkReply* reply)
{
  // Get all data from the reply
  QByteArray readData = reply->readAll();
  QString data = QString::fromUtf8(readData.constData(), readData.size());

  // Find the login url
  QRegExp formExp("loginForm\" action=\"([^\"]*)\"");
  if ( formExp.indexIn(data) != -1 )
  {
    // Set the login url
    urls.loginUrl = QUrl(formExp.cap(1));

    QRegExp urlExp("<a href=\"([^\"]*).*title=\"E-Mail\">");
    if ( urlExp.indexIn(data, formExp.pos() + formExp.matchedLength()) != -1 )
      urls.comCenterUrl = urlExp.cap(1);

    // Ok, the fields are known:
    // loginName
    // password
    // url

    QByteArray formData;
    formData += "loginName=" + loginName.toUtf8();
    formData += "&password=" + password.toUtf8();
    formData += "&url=" + urls.comCenterUrl.toString().toUtf8().toPercentEncoding();

    // Prepare a request and post it
    QNetworkRequest request = prepareRequest();
    request.setRawHeader("Referer", reply->url().toEncoded());
    request.setUrl( urls.loginUrl );

    post(request, formData, Login);
  }
  else
  {
    // Emit that we did not find the loginForm
    emit self->error( tr("loginForm could not be found!") );

    // Emit that the login failed
    emit self->loginReply(false);
  }
}

//////////////////////////////////////////////////////////////////////////

void o2GermanyPrivate::handleLogin(QNetworkReply* reply)
{
  // Get all data from the reply
  QByteArray readData = reply->readAll();
  QString data = QString::fromUtf8(readData.constData(), readData.size());

  bool success = data.indexOf("Fehler beim Login") == -1;

  if ( success )
  {
    // Open a new request to the sms center
    QNetworkRequest request = prepareRequest();
    request.setRawHeader("Referer", reply->url().toEncoded());
    request.setUrl( urls.comCenterUrl );
    get(request, ComCenter);
  }

  // Emit that the login failed
  emit self->loginReply( success );
}

//////////////////////////////////////////////////////////////////////////

void o2GermanyPrivate::handleComCenter(QNetworkReply* reply)
{
  // Get all data from the reply
  QByteArray readData = reply->readAll();

  QNetworkRequest request = prepareRequest();
  request.setRawHeader("Referer", reply->url().toEncoded());
  request.setUrl( urls.smsCenterUrl );
  get(request, SmsCenter);
}

//////////////////////////////////////////////////////////////////////////

void o2GermanyPrivate::handleSmsCenter(QNetworkReply* reply)
{
  // Get all data from the reply
  QByteArray readData = reply->readAll();
  QString data = QString::fromUtf8(readData.constData(), readData.size());

  int formOffset = 0;

  // Find the message which tells us something about our remaining free-sms
  // If it's not found we don't care ...
  QRegExp freeSmsExp("Frei-SMS: (\\d*) Web2SMS");
  if ( freeSmsExp.indexIn(data) != -1 )
  {
    // Remember the offset to not search from the beginning every time
    formOffset = freeSmsExp.pos() + freeSmsExp.matchedLength();

    // How many free sms are left?
    freeSMSCount = freeSmsExp.cap(1).toInt();

    // Emit how many free-sms we have left
    emit self->freeSmsCountChanged( freeSMSCount );
  }

  // Try to determine how many free sms we have left
  QRegExp formExp("name=\"frmSMS\" action=\"([^\"]*)\"");
  if ( formExp.indexIn(data, formOffset) != -1 )
  {
    // Remember the offset
    int offset = formExp.pos() + formExp.matchedLength();

    // Change the free sms center send-url if wanted
    urls.smsCenterSendUrl = reply->url().resolved(formExp.cap(1));

    // Fill the default fields
    sendFormDefaultValues.clear();

    QRegExp fieldExp("<input type=\"hidden\" name=\"([^\"]*)\" value=\"([^\"]*)\"", Qt::CaseInsensitive);
    int pos = fieldExp.indexIn(data, offset);
    while ( pos != -1 )
    {
      sendFormDefaultValues.insert( fieldExp.cap(1), fieldExp.cap(2).toUtf8().toPercentEncoding() );
      
      // Find the next entry
      pos = fieldExp.indexIn(data, pos+fieldExp.matchedLength());
    }

    sendFormDefaultValues.insert("Frequency", "5");
    sendFormDefaultValues.insert("StartDateDay", QByteArray());
    sendFormDefaultValues.insert("StartDateMonth", QByteArray());
    sendFormDefaultValues.insert("StartDateYear", QByteArray());
    sendFormDefaultValues.insert("StartDateHour", QByteArray());
    sendFormDefaultValues.insert("StartDateMin", QByteArray());
    sendFormDefaultValues.insert("EndDateDay", QByteArray());
    sendFormDefaultValues.insert("EndDateMonth", QByteArray());
    sendFormDefaultValues.insert("EndDateYear", QByteArray());
    sendFormDefaultValues.insert("EndDateHour", QByteArray());
    sendFormDefaultValues.insert("EndDateMin", QByteArray());
  }
  else
  {
#ifdef Q_WS_MAEMO_5
    QFile file("/home/user/MyDocs/o2sms_debug_handleSmsCenter.html");
    if ( file.open(QIODevice::Text | QIODevice::WriteOnly) )
    {
      file.write(readData);
      file.close();
    }
#endif // Q_WS_MAEMO5

    // Emit that we could not find the frmSMS
    emit self->error( tr("frmSMS could not be found!") );
  }
}

//////////////////////////////////////////////////////////////////////////

void o2GermanyPrivate::handleSmsCenterSend(QNetworkReply* reply)
{
  // Get all data from the reply
  QByteArray readData = reply->readAll();
  QString data = QString::fromUtf8(readData.constData(), readData.size());

  // TODO: ERROR CHECKING!!!!

  // Emit a signal that we've sent the message
  emit self->sendMessageReply(true, sentMessages);

  // Emit a signal that the free sms count may have changed?
  emit self->freeSmsCountChanged( --freeSMSCount );
}

//////////////////////////////////////////////////////////////////////////

void o2GermanyPrivate::handleLogout(QNetworkReply* reply)
{
  Q_UNUSED(reply);

  // No more free sms
  freeSMSCount = -1;

  // No need to read something, logout works always
  emit self->logoutReply(true);
}

//////////////////////////////////////////////////////////////////////////

void o2GermanyPrivate::onReplyFinished(QNetworkReply* reply)
{
  // Handle a redirection if needed
  QVariant redirectLocation = reply->attribute(QNetworkRequest::RedirectionTargetAttribute);
  if ( !redirectLocation.isNull() )
  {
    // Where are we going to?
    QUrl newUrl = reply->url().resolved(redirectLocation.toUrl());

    // Prepare the request and send it
    QNetworkRequest request = prepareRequest();
    request.setRawHeader("Referer", reply->url().toEncoded());
    request.setUrl( newUrl );
    get(request, replyStateHash.value(reply));

    return; // Don't continue
  }

  // If some error happend while transmitting HTTP content, also signalize it
  if ( reply->error() )
    emit self->error( tr("HTTP error happened:\n%1").arg(reply->errorString()) );

  // Decide what to do now
  switch(replyStateHash.value(reply))
  {
  case PreLogin:
    handlePreLogin(reply);
    break;

  case Login:
    handleLogin(reply);
    break;

  case ComCenter:
    handleComCenter(reply);
    break;

  case SmsCenter:
    handleSmsCenter(reply);
    break;

  case SmsCenterSend:
    handleSmsCenterSend(reply);
    break;

  case Logout:
    handleLogout(reply);
    break;

  default:
    break;
  }

  // Remove hash values for this reply (if any)
  replyStateHash.remove(reply);

  // Delete the reply at a later time
  reply->deleteLater();
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

o2Germany::o2Germany(QObject* parent /* = 0 */)
: ProviderInterface(parent)
, d(new o2GermanyPrivate)
{
  d->self = this;
}

//////////////////////////////////////////////////////////////////////////

o2Germany::~o2Germany()
{
  delete d;
}

//////////////////////////////////////////////////////////////////////////

void o2Germany::loadSettings(const QByteArray& element)
{
  QDataStream stream(element);
  stream >> d->loginName;
  stream >> d->password;
}

//////////////////////////////////////////////////////////////////////////

QByteArray o2Germany::saveSettings() const
{
  QByteArray data;
  QDataStream stream(&data, QIODevice::WriteOnly);
  stream << d->loginName;
  stream << d->password;

  return data;
}

//////////////////////////////////////////////////////////////////////////

bool o2Germany::showAccountSettingsDialog(QWidget* parent)
{
  AccountSettingsDialog dialog(parent);
  dialog.setWindowTitle( tr("o2 Germany") );
  dialog.setUserName(d->loginName);
  dialog.setPassword(d->password);

  if ( dialog.exec() == QDialog::Accepted )
  {
    d->loginName = dialog.userName();
    d->password  = dialog.password();

    return true;
  }

  return false;
}

//////////////////////////////////////////////////////////////////////////

QList<QAction*> o2Germany::actions() const
{
  return QList<QAction*>();
}

//////////////////////////////////////////////////////////////////////////

bool o2Germany::hasLoginData() const
{
  return !d->loginName.isEmpty();
}

//////////////////////////////////////////////////////////////////////////

bool o2Germany::isLoggedIn() const
{
  // Simple .. heh?
  return d->freeSMSCount != -1;
}

//////////////////////////////////////////////////////////////////////////

int o2Germany::freeSMSCount() const
{
  return d->freeSMSCount;
}

//////////////////////////////////////////////////////////////////////////

void o2Germany::login()
{
  // Start the first request
  QNetworkRequest request = d->prepareRequest();
  request.setUrl( d->urls.preLoginUrl );

  // Start the pre-login
  d->get(request, o2GermanyPrivate::PreLogin);
}

//////////////////////////////////////////////////////////////////////////

void o2Germany::logout()
{
  // Start the logout request
  QNetworkRequest request = d->prepareRequest();
  request.setUrl( d->urls.logoutUrl );

  // Start the logout request
  d->get(request, o2GermanyPrivate::Logout);
}

//////////////////////////////////////////////////////////////////////////

int o2Germany::sendMessage(const QList<ContactInfo>& receivers, const QString& text)
{
  QString smsTo;

  // Merge the numbers
  foreach(const ContactInfo& receiver, receivers)
  {
    if ( !smsTo.isEmpty() )
      smsTo += ", ";

    smsTo += QString("%1 <%2>").arg(receiver.name()).arg(receiver.number());
  }

  // Prepare the send-message request
  QNetworkRequest request = d->prepareRequest();
  request.setRawHeader("Referer", d->urls.smsCenterUrl.toEncoded());
  request.setHeader(QNetworkRequest::ContentTypeHeader, "application/x-www-form-urlencoded");
  request.setUrl( d->urls.smsCenterSendUrl );

  qDebug() << d->urls.smsCenterSendUrl;

  // Build the form data
  QByteArray formData;

  // Fill the remaining fields
  d->sendFormDefaultValues.insert("SMSTo", smsTo.toUtf8().toPercentEncoding());
  d->sendFormDefaultValues.insert("SMSText", text.toUtf8().toPercentEncoding());

  // Fill the default values
  QHashIterator<QString, QByteArray> iter(d->sendFormDefaultValues);
  while (iter.hasNext())
  {
    iter.next();

    if ( !formData.isEmpty() )
      formData += '&';

    formData += iter.key().toUtf8() + '=' + iter.value();
  }

  d->post(request, formData, o2GermanyPrivate::SmsCenterSend);

  // Return a new message id
  return d->sentMessages++;
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

class o2GermanyProviderPlugin : public QObject
                              , public ProviderPlugin
{
  Q_OBJECT
  Q_INTERFACES(ProviderPlugin)

public:
  o2GermanyProviderPlugin()
  {
    infoList << ProviderInfo(tr("o2 Germany"), "o2_germany", QIcon(":/providers/o2_germany/o2_germany.png"), QLocale::Germany);
  }

  QList<ProviderInfo> info() const
  {
    return infoList;
  }

  ProviderInterfacePtr createProvider(const QString& name) const
  {
    if ( name == "o2_germany" )
      return ProviderInterfacePtr(new o2Germany);
    
    return ProviderInterfacePtr();
  }

protected:
  QList<ProviderInfo> infoList;
};

//////////////////////////////////////////////////////////////////////////

Q_EXPORT_PLUGIN2(o2_germany, o2GermanyProviderPlugin);

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

#include "o2_germany.moc"
